import 'dart:convert';

import 'package:mysql1/mysql1.dart';

import '../date_time_extension.dart';
import '../query.dart';
import './sql.dart';
import '../attribute.dart';
import '../database.dart';
import '../document.dart';

class MariaDB extends SQL {
  final MySqlConnection _connection;

  MariaDB._(this._connection);

  Future close() async {
    await _connection.close();
  }

  static Future<MariaDB> init({
    required String host,
    required int port,
    String? user,
    String? password,
  }) async {
    final connection = await MySqlConnection.connect(ConnectionSettings(
      host: host,
      port: port,
      user: user,
      password: password,
    ));
    return MariaDB._(connection);
  }

  @override
  Future<bool> create(String name) async {
    name = filter(name);
    try {
      await _connection.query(
          'CREATE DATABASE IF NOT EXISTS `$name` /*!40100 DEFAULT CHARACTER SET utf8mb4 */;');
      return true;
    } catch (e) {
      rethrow;
    }
  }

  @override
  Future<bool> delete(String name) async {
    name = filter(name);

    final result = await _connection.query("DROP DATABASE `$name`;");

    return (result.affectedRows ?? 0) >= 1;
  }

  @override
  Future<bool> createCollection(
      String name, List<Attribute> attributes, List indexes) async {
    final id = filter(name);

    final attributeStrings = <String>[];
    final indexStrings = <String>[];

    for (var i = 0; i < attributes.length; i++) {
      final attribute = attributes[i];
      final attrId = filter(attribute.id);
      var attrType = _getSQLType(
          attribute.type, attribute.size ?? 0, attribute.signed ?? true);

      if (attribute.array ?? false) {
        attrType = 'LONGTEXT';
      }

      attributeStrings.add('`$attrId` $attrType');
    }

    for (var i = 0; i < indexes.length; i++) {
      final index = indexes[i];
      final indexId = filter(index.id);
      final indexType = index['type'];
      final indexAttributes = <String>[];

      for (var j = 0; j < index.attributes.length; j++) {
        final attribute = filter(index.attributes[j]);
        final indexLength = index.lengths?[j] ?? '';
        final indexOrder = index.orders?[j] ?? '';
        final length =
            indexLength.isNotEmpty ? '(${int.parse(indexLength)})' : '';
        final order = indexType == Database.indexFulltext ? '' : ' $indexOrder';
        indexAttributes.add('`$attribute`$length$order');
      }

      indexStrings.add('$indexType `$indexId` (${indexAttributes.join(', ')})');
    }

    try {
      final query = '''
        CREATE TABLE IF NOT EXISTS `$defaultDatabase`.`${namespace}_$id` (
          `_id` int(11) unsigned NOT NULL AUTO_INCREMENT,
          `_uid` CHAR(255) NOT NULL,
          `_createdAt` datetime(3) DEFAULT NULL,
          `_updatedAt` datetime(3) DEFAULT NULL,
          ${attributeStrings.join(',\n')},
          PRIMARY KEY (`_id`),
          ${indexStrings.join(',\n')}${indexStrings.isNotEmpty ? ',' : ''}
          UNIQUE KEY `_uid` (`_uid`),
          KEY `_created_at` (`_createdAt`),
          KEY `_updated_at` (`_updatedAt`)
        )
      ''';
      await _connection.query(query);

      return true;
    } catch (e) {
      // await _handleError(e, id);
      return false;
    }
  }

  @override
  Future<bool> deleteCollection(String name) async {
    try {
      await _connection.query("DROP TABLE ${getSQLTable(name)};");
      return true;
    } on MySqlException catch (_) {
      return false;
    }
  }

  @override
  Future<bool> createAttribute(String collection, Attribute attribute) async {
    final name = filter(collection);
    final id = filter(attribute.id);
    var sqlType = _getSQLType(
        attribute.type, attribute.size ?? 0, attribute.signed ?? true);

    if (attribute.array ?? false) {
      sqlType = 'LONGTEXT';
    }

    try {
      await _connection
          .query("ALTER TABLE ${getSQLTable(name)} ADD COLUMN `$id` $sqlType;");
      return true;
    } on MySqlException catch (_) {
      return false;
    }
  }

  @override
  Future<bool> updateAttribute(String collection, Attribute attribute) async {
    final name = filter(collection);
    final id = filter(attribute.id);
    var sqlType = _getSQLType(
        attribute.type, attribute.size ?? 0, attribute.signed ?? true);

    if (attribute.array ?? false) {
      sqlType = 'LONGTEXT';
    }
    final query = "ALTER TABLE ${getSQLTable(name)} MODIFY `$id` $sqlType;";
    try {
      await _connection.query(query);
      return true;
    } on MySqlException catch (_) {
      return false;
    }
  }

  @override
  Future<bool> deleteAttribute(String collection, String id) async {
    final name = filter(collection);
    id = filter(id);

    try {
      await _connection
          .query("ALTER TABLE ${getSQLTable(name)} DROP COLUMN `$id`;");
      return true;
    } on MySqlException catch (_) {
      return false;
    }
  }

  @override
  Future<bool> renameAttribute(
      String collection, String old, String newName) async {
    final name = filter(collection);
    old = filter(old);
    newName = filter(newName);

    try {
      await _connection.query(
          "ALTER TABLE ${getSQLTable(name)} RENAME COLUMN `$old` TO `$newName`;");
      return true;
    } on MySqlException catch (_) {
      return false;
    }
  }

  @override
  Future<bool> createRelationship(
      String collection, String relatedCollection, String type,
      {bool twoWay = false, String id = '', String twoWayKey = ''}) async {
    String name = filter(collection);
    String relatedName = filter(relatedCollection);
    String table = getSQLTable(name);
    String relatedTable = getSQLTable(relatedName);
    id = filter(id);
    twoWayKey = filter(twoWayKey);
    String sqlType = _getSQLType(Database.varRelationship, 0, false);

    String sql = '';
    switch (type) {
      case Database.relationOneToOne:
        sql = "ALTER TABLE $table ADD COLUMN `$id` $sqlType DEFAULT NULL;";
        if (twoWay) {
          sql +=
              "ALTER TABLE $relatedTable ADD COLUMN `$twoWayKey` $sqlType DEFAULT NULL;";
        }
        break;
      case Database.relationOneToMany:
        sql =
            "ALTER TABLE $relatedTable ADD COLUMN `$twoWayKey` $sqlType DEFAULT NULL;";
        break;
      case Database.relationManyToOne:
        sql = "ALTER TABLE $table ADD COLUMN `$id` $sqlType DEFAULT NULL;";
        break;
      case Database.relationManyToMany:
        return true;
      default:
        throw Exception('Invalid relationship type.');
    }

    var result = await _connection.query(sql);
    return (result.affectedRows ?? 0) > 0;
  }

  @override
  Future<bool> updateRelationship(String collection, String relatedCollection,
      String type, bool twoWay, String key, String twoWayKey,
      [String? newKey, String? newTwoWayKey]) async {
    var name = filter(collection);
    var relatedName = filter(relatedCollection);
    var table = getSQLTable(name);
    var relatedTable = getSQLTable(relatedName);
    key = filter(key);
    twoWayKey = filter(twoWayKey);

    if (newKey != null) {
      newKey = filter(newKey);
    }
    if (newTwoWayKey != null) {
      newTwoWayKey = filter(newTwoWayKey);
    }

    var sql = '';

    switch (type) {
      case Database.relationOneToOne:
        if (newKey != null) {
          sql = "ALTER TABLE $table RENAME COLUMN $key TO $newKey;";
        }
        if (twoWay && newTwoWayKey != null) {
          sql +=
              "ALTER TABLE $relatedTable RENAME COLUMN $twoWayKey TO $newTwoWayKey;";
        }
        break;
      case Database.relationOneToMany:
        if (twoWay && newTwoWayKey != null) {
          sql =
              "ALTER TABLE $relatedTable RENAME COLUMN $twoWayKey TO $newTwoWayKey;";
        }
        break;
      case Database.relationManyToOne:
        if (newKey != null) {
          sql = "ALTER TABLE $table RENAME COLUMN $key TO $newKey;";
        }
        break;
      case Database.relationManyToMany:
        var collectionDoc = await getDocument(Database.metadata, collection);
        var relatedCollectionDoc =
            await getDocument(Database.metadata, relatedCollection);
        var junction = getSQLTable(
            '${collectionDoc.internalId}${relatedCollectionDoc.internalId}');

        if (newKey != null) {
          sql = "ALTER TABLE $junction RENAME COLUMN `$key` TO `$newKey`;";
        }
        if (twoWay && newTwoWayKey != null) {
          sql +=
              "ALTER TABLE $junction RENAME COLUMN `$twoWayKey` TO `$newTwoWayKey`;";
        }
        break;
      default:
        throw Exception('Invalid relationship type.');
    }

    if (sql.isEmpty) {
      return true;
    }

    var results = await _connection.query(sql);
    return (results.affectedRows ?? 0) > 0;
  }

  @override
  Future<bool> deleteRelationship(
    String collection,
    String relatedCollection,
    String type,
    bool twoWay,
    String key,
    String twoWayKey,
    String side,
  ) async {
    final name = filter(collection);
    final relatedName = filter(relatedCollection);
    final table = getSQLTable(name);
    final relatedTable = getSQLTable(relatedName);
    key = filter(key);
    twoWayKey = filter(twoWayKey);

    var sql = '';

    switch (type) {
      case Database.relationOneToOne:
        sql = "ALTER TABLE $table DROP COLUMN '$key';";
        if (twoWay) {
          sql += "ALTER TABLE $relatedTable DROP COLUMN '$twoWayKey';";
        }
        break;
      case Database.relationOneToMany:
        if (side == Database.relationSideParent) {
          sql = "ALTER TABLE $relatedTable DROP COLUMN '$twoWayKey';";
        } else if (twoWay) {
          sql = "ALTER TABLE $table DROP COLUMN '$key';";
        }
        break;
      case Database.relationManyToOne:
        if (twoWay && side == Database.relationSideChild) {
          sql = "ALTER TABLE $relatedTable DROP COLUMN '$twoWayKey';";
        } else {
          sql = "ALTER TABLE $table DROP COLUMN '$key';";
        }
        break;
      case Database.relationManyToMany:
        final collectionMeta = await getDocument(Database.metadata, collection);
        final relatedCollectionMeta =
            await getDocument(Database.metadata, relatedCollection);

        final junction = side == Database.relationSideParent
            ? getSQLTable(
                '_${collectionMeta.internalId}_${relatedCollectionMeta.internalId}')
            : getSQLTable(
                '_${relatedCollectionMeta.internalId}_${collectionMeta.internalId}');

        final perms = side == Database.relationSideParent
            ? getSQLTable(
                '_${collectionMeta.internalId}_${relatedCollectionMeta.internalId}_perms')
            : getSQLTable(
                '_${relatedCollectionMeta.internalId}_${collectionMeta.internalId}_perms');

        sql = "DROP TABLE $junction; DROP TABLE $perms";
        break;
      default:
        throw Exception('Invalid relationship type.');
    }

    if (sql.isEmpty) {
      return true;
    }

    var result = await _connection.query(sql);
    return (result.affectedRows ?? 0) > 0;
  }

  @override
  Future<Document> createDocument(String collection, Document document) async {
    final attributes = document.getAttributes();
    attributes['_createdAt'] = (document.createdAt ?? DateTime.now()).format();
    attributes['_updatedAt'] = (document.updatedAt ?? DateTime.now()).format();

    final name = filter(collection);
    var columns = <String>[];
    var columnNames = <String>[];
    var values = <Object>[];

    try {
      for (final attribute in attributes.entries) {
        final column = filter(attribute.key);
        columns.add('`$column`');
        columnNames.add('?');
        values.add(attribute.value is List || attribute.value is Map
            ? json.encode(attribute.value)
            : attribute.value);
      }

      final statement = 'INSERT INTO ${getSQLTable(name)} '
          '(${columns.join(', ')}, `_uid`) '
          'VALUES (${columnNames.join(', ')}, ?)';

      values.add(document.id!);

      await _connection.query(statement, values);

      final documentResult = await getDocument(
        collection,
        document.id!,
      );
      document['\$internalId'] = documentResult.internalId;
    } on MySqlException catch (e) {
      switch (e.errorNumber) {
        case 1062:
        case 23000:
          throw Exception('Duplicated document: ${e.message}');
        default:
          rethrow;
      }
    }

    return document;
  }

  @override
  Future<Document> getDocument(String collection, String id,
      {List<Query> queries = const []}) async {
    final name = filter(collection);
    final selections = getAttributeSelections(queries);

    final query = '''
      SELECT ${getAttributeProjection(selections, '')}
      FROM ${getSQLTable(name)}
      WHERE _uid = ?
    ''';
    final results = await _connection.query(query, [id]);

    if (results.isEmpty) {
      return Document({});
    }

    final document = results.first.fields;

    document['\$id'] = document['_uid'];
    document['\$internalId'] = document['_id'];
    document['\$createdAt'] = document['_createdAt'];
    document['\$updatedAt'] = document['_updatedAt'];

    document.remove('_id');
    document.remove('_uid');
    document.remove('_createdAt');
    document.remove('_updatedAt');

    return Document(document);
  }

  @override
  Future<bool> deleteDocument(String collection, String id) async {
    var name = filter(collection);

    try {
      await _connection.transaction((txn) async {
        await txn
            .query('DELETE FROM ${getSQLTable(name)} WHERE _uid = ?', [id]);
      });
      return true;
    } catch (e) {
      rethrow;
    }
  }

  @override
  Future<bool> createIndex(
      String collection, String id, String type, List<String> attributes,
      [List<int> lengths = const [], List<String> orders = const []]) async {
    final collectionDoc = await getDocument(Database.metadata, collection);
    if (collectionDoc.isEmpty) {
      throw Exception('Collection not found');
    }

    id = filter(id);
    for (var i = 0; i < attributes.length; i++) {
      var attr = attributes[i];
      var order = (orders.isEmpty ||
              orders[i].isEmpty ||
              type == Database.indexFulltext)
          ? ''
          : orders[i];
      var length =
          (lengths.isEmpty || lengths[i] == 0) ? '' : '(${lengths[i]})';

      switch (attr) {
        case '\$id':
          attr = '_uid';
          break;
        case '\$createdAt':
          attr = '_createdAt';
          break;
        case '\$updatedAt':
          attr = '_updatedAt';
          break;
        default:
          attr = filter(attr);
      }

      attributes[i] = "`$attr`$length $order";
    }

    var sqlType = '';
    switch (type) {
      case Database.indexKey:
        sqlType = 'INDEX';
        break;
      case Database.indexUnique:
        sqlType = 'UNIQUE INDEX';
        break;
      case Database.indexFulltext:
        sqlType = 'FULLTEXT INDEX';
        break;
      default:
        throw Exception(
            'Unknown index type: $type. Must be one of INDEX, UNIQUE INDEX, FULLTEXT INDEX');
    }

    var joinedAttributes = attributes.join(', ');

    var sql =
        "CREATE $sqlType `$id` ON ${getSQLTable(collectionDoc.id!)} ($joinedAttributes)";
    try {
      await _connection.query(sql);
      return true;
    } catch (e) {
      return false;
    }
  }

  @override
  Future<bool> renameIndex(
      String collection, String old, String newName) async {
    collection = filter(collection);
    old = filter(old);
    newName = filter(newName);

    final query =
        'ALTER TABLE ${getSQLTable(collection)} RENAME INDEX $old TO $newName;';
    try {
      await _connection.query(query);
      return true;
    } catch (e) {
      return false;
    }
  }

  @override
  Future<bool> deleteIndex(String collection, String id) async {
    collection = filter(collection);
    id = filter(id);
    final query = 'ALTER TABLE ${getSQLTable(collection)} DROP INDEX `$id`;';
    await _connection.query(query);
    return true;
  }

  String getAttributeProjection(List<String> selections, String prefix) {
    if (selections.isEmpty || selections.contains('')) {
      if (prefix.isNotEmpty) {
        return "$prefix.*";
      }
      return '*';
    }

    selections.removeWhere((element) => [
          '\$id',
          '\$collection',
          '\$internalId',
          '\$createdAt',
          '\$updatedAt'
        ].contains(element));

    selections.addAll(['_uid', '_id', '_createdAt', '_updatedAt']);

    if (prefix.isNotEmpty) {
      for (var i = 0; i < selections.length; i++) {
        selections[i] = "$prefix.${selections[i]}";
      }
    } else {
      for (var i = 0; i < selections.length; i++) {
        selections[i] = selections[i];
      }
    }

    return selections.join(', ');
  }

  String getSQLSchema() {
    if (!getSupportForSchemas()) {
      return '';
    }
    return '`${getDefaultDatabase()}`.';
  }

  String getSQLTable(String name) {
    return '${getSQLSchema()}`${getNamespace()}_$name`';
  }

  bool getSupportForSchemas() {
    return true;
  }

  @override
  Future<bool> ping() async {
    try {
      await _connection.query('SELECT 1;');
      return true;
    } catch (e) {
      return false;
    }
  }

  @override
  Future<bool> exists(String database, {String? collection}) async {
    database = filter(database);

    late String select, from, where, match;
    if (collection != null) {
      collection = filter(collection);

      select = 'TABLE_NAME';
      from = 'INFORMATION_SCHEMA.TABLES';
      where = 'TABLE_SCHEMA = ? AND TABLE_NAME = ?';
      match = '${getNamespace()}_$collection';
    } else {
      select = 'SCHEMA_NAME';
      from = 'INFORMATION_SCHEMA.SCHEMATA';
      where = 'SCHEMA_NAME = ?';
      match = database;
    }

    final query = 'SELECT $select FROM $from WHERE $where';
    final values = collection != null
        ? [database, '${getNamespace()}_$collection']
        : [database];
    final results = await _connection.query(
      query,
      values,
    );

    if (results.isEmpty) return false;
    final document = results.first.fields;

    return (document[select] ?? '') == match ||
        (document[select.toLowerCase()] ?? '') == match;
  }

  @override
  Future<List> list() {
    throw UnimplementedError();
  }

  String _getSQLType(String type, int size,
      [bool isSigned = true, bool array = false]) {
    if (array) {
      return 'JSON';
    }

    String signed = isSigned ? '' : ' UNSIGNED';
    switch (type) {
      case Database.varString:
        // size = size * 4; // Convert utf8mb4 size to bytes
        if (size > 16777215) {
          return 'LONGTEXT';
        }

        if (size > 65535) {
          return 'MEDIUMTEXT';
        }

        if (size > 16383) {
          return 'TEXT';
        }

        return "VARCHAR($size)";

      case Database.varInteger:
        // We don't support zerofill: https://stackoverflow.com/a/5634147/2299554

        if (size >= 8) {
          // INT = 4 bytes, BIGINT = 8 bytes
          return 'BIGINT$signed';
        }

        return 'INT$signed';

      case Database.varFloat:
        return 'DOUBLE$signed';

      case Database.varBoolean:
        return 'TINYINT(1)';

      case Database.varRelationship:
        return 'VARCHAR(255)';

      case Database.varDateTime:
        return 'DATETIME(3)';

      default:
        throw Exception('Unknown Type');
    }
  }

  @override
  bool getSupportForCasting() {
    return false;
  }

  @override
  Future<bool> updateDocument(String collection, Document document) async {
    final name = filter(collection);
    final attributes = document.getAttributes();
    attributes['_createdAt'] = document.createdAt?.format();
    attributes['_updatedAt'] = document.updatedAt?.format();

    var columns = '';

    var updateSql = "UPDATE ${getSQLTable(name)} SET ";
    var updateValues = <dynamic>[];
    attributes.forEach((attribute, value) {
      var sanitizedAttribute = filter(attribute);
      columns += "`$sanitizedAttribute` = ?, ";
      updateValues.add(value);
    });

    updateSql = "$updateSql${columns}_uid = ? WHERE _uid = ?";
    updateValues.add(document.id!);
    updateValues.add(document.id!);

    try {
      await _connection.query(updateSql, updateValues);
      return true;
    } catch (e) {
      return false;
    }
  }

  @override
  Future<List<Document>> find(
    String collectionId, {
    List<Query> queries = const [],
    int limit = 25,
    int? offset,
    List<String> orderAttributes = const [],
    List<String> orderTypes = const [],
    Map<String, dynamic> cursor = const {},
    String cursorDirection = Database.cursorAfter,
  }) async {
    var name = filter(collectionId);
    var where = <String>[];
    var orders = <String>[];

    var hasIdAttribute = false;
    for (var i = 0; i < orderAttributes.length; i++) {
      var attribute =
          orderAttributes[i] == '\$id' ? '_uid' : orderAttributes[i];
      if (attribute == '_uid') {
        hasIdAttribute = true;
      }

      var orderType = orderTypes[i];
      if (cursorDirection == 'BEFORE') {
        orderType = orderType == 'ASC' ? 'DESC' : 'ASC';
      }

      orders.add('`${filter(attribute)}` $orderType');
    }

    if (orders.isEmpty && cursor.isNotEmpty) {
      var orderType =
          orderTypes.isNotEmpty ? orderTypes.first : Database.orderAsc;
      var orderMethod = Query.typeLesser;
      if (cursorDirection == Database.cursorAfter) {
        orderMethod = orderType == Database.orderDesc
            ? Query.typeLesser
            : Query.typeGreater;
      } else {
        orderMethod = orderType == Database.orderDesc
            ? Query.typeGreater
            : Query.typeLesser;
      }
      where.add(
          '( table_main._id ${getSQLOperator(orderMethod)} ${cursor["\$internalId"]})');
    }

    if (!hasIdAttribute) {
      if (orderAttributes.isEmpty && orderTypes.isNotEmpty) {
        var order = orderTypes.first;
        if (cursorDirection == Database.cursorBefore) {
          order = order == Database.orderAsc
              ? Database.orderDesc
              : Database.orderAsc;
        }

        orders.add('table_main._id ${filter(order)}');
      } else {
        orders.add(
            'table_main._id ${cursorDirection == Database.cursorAfter ? Database.orderAsc : Database.orderDesc}');
      }
    }

    var conditions = getSQLConditions(queries);
    if (conditions.isNotEmpty) {
      where.add(conditions);
    }

    var sqlWhere = where.isNotEmpty ? 'WHERE ${where.join(' AND ')}' : '';
    var sqlOrder = 'ORDER BY ${orders.join(', ')}';
    var sqlLimit = 'LIMIT $limit';
    sqlLimit += offset != null ? ' OFFSET $offset' : '';

    var selections = getAttributeSelections(queries);

    var parameters = [];
    var cursorCondition = '';
    if (cursor.isNotEmpty && orderAttributes.isNotEmpty) {
      var cursorAttribute = orderAttributes.first;
      var cursorValue = cursor[cursorAttribute];
      if (cursorValue == null) {
        throw Exception(
            "Cursor value for attribute '$cursorAttribute' is empty");
      }
      cursorCondition = " AND `${filter(cursorAttribute)}` > ?";
      parameters.add(cursorValue);
    }

    var sql = '''
      SELECT ${getAttributeProjection(selections, 'table_main')}
      FROM ${getSQLTable(name)} as table_main
      $sqlWhere
      $cursorCondition
      $sqlOrder
      $sqlLimit;
    ''';
    print(sql);
    final results = await _connection.query(sql);

    List<Document> documents = results.map((res) {
      final document = Document(res.fields);
      if (document.containsKey('_uid')) {
        document['\$id'] = document['_uid'];
        document.remove('_uid');
      }
      if (document.containsKey('_id')) {
        document['\$internalId'] = document['_id'];
        document.remove('_id');
      }
      if (document.containsKey('_tenant')) {
        document['\$tenant'] = document['_tenant'];
        document.remove('_tenant');
      }
      if (document.containsKey('_createdAt')) {
        document['\$createdAt'] = document['_createdAt'];
        document.remove('_createdAt');
      }
      if (document.containsKey('_updatedAt')) {
        document['\$updatedAt'] = document['_updatedAt'];
        document.remove('_updatedAt');
      }
      if (document.containsKey('_permissions')) {
        document['\$permissions'] =
            jsonDecode(document['_permissions'] ?? '[]');
        document.remove('_permissions');
      }
      return document;
    }).toList();

    if (cursorDirection == Database.cursorBefore) {
      documents.reversed.toList();
    }

    return documents;
  }

  String getSQLOperator(String method) {
    switch (method) {
      case Query.typeEqual:
        return '=';
      case Query.typeNotEqual:
        return '!=';
      case Query.typeLesser:
        return '<';
      case Query.typeLesserEqual:
        return '<=';
      case Query.typeGreater:
        return '>';
      case Query.typeGreaterEqual:
        return '>=';
      case Query.typeIsNull:
        return 'IS NULL';
      case Query.typeIsNotNull:
        return 'IS NOT NULL';
      case Query.typeStartsWith:
      case Query.typeEndsWith:
      case Query.typeContains:
        // Assuming getLikeOperator() logic needs to be integrated or replaced
        return 'LIKE';
      default:
        throw Exception('Unknown method: $method');
    }
  }

  @override
  String getSQLCondition(Query query) {
    query = query.copyWith(
        attribute: {
      '\$id': '_uid',
      '\$internalId': '_id',
      '\$tenant': '_tenant',
      '\$createdAt': '_createdAt',
      '\$updatedAt': '_updatedAt',
    }[query.attribute]);

    var attribute = "`${query.attribute}`";

    switch (query.method) {
      case Query.typeOr:
      case Query.typeAnd:
        var conditions = query.values.map((q) => getSQLCondition(q)).toList();
        var method = query.method.toUpperCase();
        return conditions.isEmpty
            ? ''
            : ' $method (${conditions.join(' AND ')})';
      case Query.typeSearch:
        return "MATCH(`table_main`.$attribute) AGAINST (${query.values[0]} IN BOOLEAN MODE)";
      case Query.typeBetween:
        return "`table_main`.$attribute BETWEEN '${query.values[0]}' AND '${query.values[1]}'";
      case Query.typeIsNull:
      case Query.typeIsNotNull:
        return "`table_main`.$attribute ${getSQLOperator(query.method)}";
      case Query.typeContains:
      // Assuming getSupportForJSONOverlaps() and query.onArray() logic needs to be integrated
      // Fallback to default for simplicity
      default:
        var conditions = query.values
            .asMap()
            .entries
            .map((entry) =>
                "$attribute ${getSQLOperator(query.method)} '${entry.value}'")
            .toList();
        return conditions.isEmpty ? '' : '(${conditions.join(' OR ')})';
    }
  }
}
